"""Configuration file for the Sphinx documentation builder for pybticino.

This file configures Sphinx to build the documentation for the pybticino
library, including settings for project information, extensions, and HTML output.
"""

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from pathlib import Path  # Import Path
import sys

# Point to the source code directory relative to the docs directory
# Use pathlib instead
src_path = Path(__file__).parent.parent / "src"
sys.path.insert(0, str(src_path.resolve()))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pybticino"
copyright = "2025, k-the-hidden-hero"  # Replace with actual year/author if different
author = "k-the-hidden-hero"  # Replace with actual author if different

# Attempt to get version from __init__.py - replace with a fixed version if needed
try:
    from pybticino import __version__ as release
except ImportError:
    release = "0.1.0"  # Fallback version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",  # Include documentation from docstrings
    "sphinx.ext.napoleon",  # Support for Google and NumPy style docstrings
    "sphinx.ext.viewcode",  # Add links to source code
    "sphinx.ext.intersphinx",  # Link to other projects' documentation
    "myst_parser",  # Allow Markdown files
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"  # Read the Docs theme

html_static_path = ["_static"]

# -- Options for autodoc -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

autodoc_member_order = "bysource"  # Order members by source code order
autodoc_default_options = {
    "members": True,
    "undoc-members": True,  # Include members without docstrings (though we added them)
    "show-inheritance": True,
}

# -- Options for intersphinx extension ---------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "aiohttp": ("https://docs.aiohttp.org/en/stable/", None),
    "websockets": ("https://websockets.readthedocs.io/en/stable/", None),
}

# -- Options for Napoleon extension ------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html#configuration

napoleon_google_docstring = True
napoleon_numpy_docstring = False  # Set to True if you use NumPy style
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True



"""Shared fixtures for pybticino tests."""

import pytest

from pybticino import AuthHandler
from pybticino.const import BASE_URL, GETEVENTS_ENDPOINT, HOMESDATA_ENDPOINT, HOMESTATUS_ENDPOINT, SETSTATE_ENDPOINT

# --- Test data ---

MOCK_USERNAME = "user@example.com"
MOCK_PASSWORD = "secret123"
MOCK_ACCESS_TOKEN = "mock-access-token-abc123"
MOCK_REFRESH_TOKEN = "mock-refresh-token-xyz789"
MOCK_HOME_ID = "home_test_001"
MOCK_HOME_NAME = "Casa Test"
MOCK_BRIDGE_ID = "AA:BB:CC:DD:EE:FF"
MOCK_LOCK_ID = "module_lock_1"
MOCK_LIGHT_ID = "module_light_1"

TOKEN_URL = BASE_URL + "/oauth2/token"
HOMESDATA_URL = BASE_URL + HOMESDATA_ENDPOINT
HOMESTATUS_URL = BASE_URL + HOMESTATUS_ENDPOINT
SETSTATE_URL = BASE_URL + SETSTATE_ENDPOINT
GETEVENTS_URL = BASE_URL + GETEVENTS_ENDPOINT


def build_token_response(
    access_token=MOCK_ACCESS_TOKEN,
    refresh_token=MOCK_REFRESH_TOKEN,
    expires_in=3600,
):
    """Build a mock OAuth2 token response."""
    return {
        "access_token": access_token,
        "refresh_token": refresh_token,
        "expires_in": expires_in,
        "token_type": "Bearer",
        "scope": "security_scopes",
    }


def build_homesdata_response():
    """Build a mock /api/homesdata response."""
    return {
        "body": {
            "user": {"email": MOCK_USERNAME},
            "homes": [
                {
                    "id": MOCK_HOME_ID,
                    "name": MOCK_HOME_NAME,
                    "modules": [
                        {
                            "id": MOCK_BRIDGE_ID,
                            "name": "Bridge",
                            "type": "BNCX",
                            "firmware_name": "2.1.0",
                        },
                        {
                            "id": MOCK_LOCK_ID,
                            "name": "Front Door Lock",
                            "type": "BNDL",
                            "variant": "xxx:bndl_doorlock",
                            "bridge": MOCK_BRIDGE_ID,
                        },
                        {
                            "id": MOCK_LIGHT_ID,
                            "name": "Staircase Light",
                            "type": "BNSL",
                            "variant": "xxx:bnsl_staircase_light",
                            "bridge": MOCK_BRIDGE_ID,
                        },
                    ],
                },
            ],
        },
        "status": "ok",
    }


def build_homestatus_response():
    """Build a mock /syncapi/v1/homestatus response."""
    return {
        "body": {
            "home": {
                "id": MOCK_HOME_ID,
                "modules": [
                    {"id": MOCK_LOCK_ID, "lock": True, "reachable": True},
                    {"id": MOCK_LIGHT_ID, "status": "off", "reachable": True},
                    {"id": MOCK_BRIDGE_ID, "wifi_strength": 65, "uptime": 86400},
                ],
            },
        },
        "status": "ok",
    }


def build_events_response():
    """Build a mock /api/getevents response."""
    return {
        "body": {
            "home": {
                "id": MOCK_HOME_ID,
                "events": [
                    {
                        "id": "evt_1",
                        "type": "call",
                        "module_id": "module_ext_unit_1",
                        "time": 1700000000,
                        "subevents": [
                            {
                                "type": "missed_call",
                                "time": 1700000050,
                            }
                        ],
                    },
                ],
            },
        },
        "status": "ok",
    }


@pytest.fixture
def auth_handler():
    """Create an AuthHandler instance for testing."""
    return AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)



"""Represents a pybticino account."""

from __future__ import annotations

import logging
from typing import Any

import aiohttp

# Import AuthHandler for runtime checks like isinstance
from .auth import AuthHandler
from .const import (
    BASE_URL,
    DEFAULT_ANDROID_VERSION,
    DEFAULT_APP_VERSION,
    DEFAULT_BUILD_NUMBER,
    DEFAULT_DEVICE_INFO,
    GETEVENTS_ENDPOINT,
    HOMESDATA_ENDPOINT,
    HOMESTATUS_ENDPOINT,
    SETSTATE_ENDPOINT,
    # Add other endpoints as needed
    build_user_agent,
)
from .exceptions import ApiError, AuthError
from .models import Home, Module  # Import basic models

LOG = logging.getLogger(__name__)


# Define default device types here or import from const if moved
DEFAULT_DEVICE_TYPES: list[str] = [
    "BNMH",
    "BNCX",
    "BFII",
    "BPAC",
    "BPVC",
    "BNC1",
    "BDIY",
    "BNHY",
    "NACamera",
    "NOC",
    "NDB",
    "NSD",
    "NCO",
    "NDL",
]


class AsyncAccount:
    """Represents a BTicino user account and provides methods to interact with the API.

    This class is the main entry point for interacting with the user's homes and
    modules after successful authentication. It uses an `AuthHandler` instance
    to manage authentication tokens and an `aiohttp.ClientSession` (obtained
    from the `AuthHandler`) to make API requests.

    Attributes:
        auth_handler (AuthHandler): The authentication handler instance used for
                                    managing tokens and the HTTP session.
        user (Optional[str]): The email address of the authenticated user, populated
                              after calling `async_update_topology`.
        homes (dict[str, Home]): A dictionary mapping home IDs to `Home` objects,
                                 populated after calling `async_update_topology`.
        raw_data (dict[str, Any]): The last raw response received from the
                                   `/homesdata` endpoint.

    """

    def __init__(
        self,
        auth_handler: AuthHandler,
        app_version: str = DEFAULT_APP_VERSION,
        build_number: str = DEFAULT_BUILD_NUMBER,
        android_version: str = DEFAULT_ANDROID_VERSION,
        device_info: str = DEFAULT_DEVICE_INFO,
    ) -> None:
        """Initialize the AsyncAccount.

        Args:
            auth_handler (AuthHandler): An initialized and authenticated
                                        `AuthHandler` instance.
            app_version (str): The application version string to use in User-Agent.
                               Defaults to `DEFAULT_APP_VERSION`.
            build_number (str): The build number string to use in User-Agent.
                                Defaults to `DEFAULT_BUILD_NUMBER`.
            android_version (str): The Android version string to use in User-Agent.
                                   Defaults to `DEFAULT_ANDROID_VERSION`.
            device_info (str): The device info string to use in User-Agent.
                               Defaults to `DEFAULT_DEVICE_INFO`.

        Raises:
            TypeError: If `auth_handler` is not an instance of `AuthHandler`.

        """
        # Runtime check needs AuthHandler to be defined
        if not isinstance(auth_handler, AuthHandler):
            err_msg = "auth_handler must be an instance of AuthHandler"
            raise TypeError(err_msg)
        self.auth_handler: AuthHandler = auth_handler
        self.user: str | None = None  # Store user email if available
        self.homes: dict[str, Home] = {}  # Store Home objects keyed by home_id
        self.raw_data: dict[str, Any] = {}  # Store last raw homesdata response

        # Store User-Agent components
        self._app_version = app_version
        self._build_number = build_number
        self._android_version = android_version
        self._device_info = device_info

    async def _async_post_api_request(
        self,
        endpoint: str,
        params: dict[str, Any] | None = None,
        json_data: dict[str, Any] | None = None,
        timeout: int = 30,
    ) -> dict[str, Any]:
        """Make an authenticated asynchronous POST request to the BTicino API.

        This is a helper method that handles fetching a valid access token,
        constructing headers (including Authorization and User-Agent), and
        making the POST request using the shared `aiohttp.ClientSession`.
        It also handles common API error responses.

        Args:
            endpoint (str): The API endpoint path (e.g., '/api/homesdata').
            params (Optional[dict[str, Any]]): URL parameters for the request.
            json_data (Optional[dict[str, Any]]): JSON payload for the request body.
            timeout (int): Request timeout in seconds. Defaults to 15.

        Returns:
            dict[str, Any]: The JSON response body as a dictionary. Returns an
                            empty dictionary for 204 No Content responses.

        Raises:
            AuthError: If obtaining an access token fails.
            ApiError: If the API returns an error status code (>= 400),
                      if the request times out, or if a client-side error occurs.

        """
        try:
            access_token = await self.auth_handler.get_access_token()
        except AuthError:
            LOG.exception("Authentication required but failed")
            raise

        url = BASE_URL + endpoint
        user_agent = build_user_agent(
            app_version=self._app_version,
            build_number=self._build_number,
            android_version=self._android_version,
            device_info=self._device_info,
        )
        headers = {
            "Authorization": f"Bearer {access_token}",
            "User-Agent": user_agent,
            "Content-Type": "application/json; charset=utf-8",  # Assume JSON for POST
        }

        session = await self.auth_handler._get_session()  # type: ignore[protected-access]

        LOG.debug("Making ASYNC POST request to %s", url)
        LOG.debug(
            "Headers: %s",
            {k: (v[:30] + "..." if k == "Authorization" else v) for k, v in headers.items()},
        )
        LOG.debug("Params: %s", params)
        LOG.debug("JSON Data: %s", json_data)

        try:
            async with session.post(
                url,
                headers=headers,
                params=params,
                json=json_data,
                timeout=timeout,
            ) as response:
                LOG.debug("Response status code: %s", response.status)
                LOG.debug("Response headers: %s", response.headers)

                if response.status >= 400:
                    error_text = await response.text()
                    LOG.error(
                        "API Error Response (%s): %s",
                        response.status,
                        error_text,
                    )
                    try:
                        # Try to parse specific error format
                        error_content = await response.json()
                        error_message = (
                            error_content.get("error", {}).get("message") or error_content.get("error") or error_text
                        )
                    except aiohttp.ContentTypeError, ValueError:
                        error_message = error_text
                    raise ApiError(response.status, error_message)

                # Handle empty response body for certain status codes if necessary
                if response.status == 204:  # No Content
                    return {}  # Return empty dict for consistency

                # Assuming JSON response for successful calls
                return await response.json()

        except aiohttp.ClientResponseError as http_err:
            # This might be redundant if ApiError is raised above, but acts as a fallback
            LOG.exception("HTTP error during API request")
            if not isinstance(http_err, ApiError):
                raise ApiError(http_err.status, str(http_err)) from http_err
            raise  # Re-raise the original ApiError
        except TimeoutError:
            LOG.exception("Request timed out: POST %s", url)
            raise ApiError(408, "Request timed out") from None
        except aiohttp.ClientError as req_err:
            LOG.exception("Request error during API request")
            raise ApiError(0, f"Request error: {req_err}") from req_err
        except Exception as e:
            LOG.exception("Unexpected error during API request")
            raise ApiError(0, f"Unexpected error: {e}") from e

    async def async_update_topology(
        self,
        device_types: list[str] | None = None,
    ) -> None:
        """Fetch the user's home topology (homes and modules) from the API.

        This method calls the `/api/homesdata` endpoint to retrieve information
        about the user's homes and the modules within them. It populates the
        `self.homes` dictionary with `Home` objects and `self.user` with the
        user's email address. It also stores the raw response in `self.raw_data`.

        Args:
            device_types (Optional[list[str]]): A list of device type strings to
                filter the results. Defaults to `DEFAULT_DEVICE_TYPES`.

        Raises:
            AuthError: If obtaining an access token fails.
            ApiError: If the API call fails.

        """
        if device_types is None:
            device_types = DEFAULT_DEVICE_TYPES

        payload = {
            "app_type": "app_camera",  # Keep this specific type for pybticino
            "app_version": self._app_version,
            "device_types": device_types,
            "sync_measurements": False,
        }

        homes_data = await self._async_post_api_request(
            endpoint=HOMESDATA_ENDPOINT,
            json_data=payload,
        )

        # Basic processing similar to pyatmo
        self.raw_data = homes_data  # Store the raw response
        body = self.raw_data.get("body", {})
        self.user = body.get("user", {}).get("email")

        # Clear existing homes before processing new data
        self.homes.clear()

        for home_data in body.get("homes", []):
            home_id = home_data.get("id")
            if not home_id:
                LOG.warning("Skipping home with missing ID: %s", home_data.get("name"))
                continue

            # Create basic Module objects (can be refined later)
            modules = []
            for module_data in home_data.get("modules", []):
                mod_id = module_data.get("id")
                if not mod_id:
                    LOG.warning("Skipping module with missing ID in home %s", home_id)
                    continue
                modules.append(
                    Module(
                        id=mod_id,
                        name=module_data.get("name", "Unknown Module"),
                        type=module_data.get("type", "Unknown Type"),
                        bridge=module_data.get("bridge"),
                        raw_data=module_data,
                    ),
                )

            # Create Home object
            self.homes[home_id] = Home(
                id=home_id,
                name=home_data.get("name", "Unknown Home"),
                modules=modules,
                raw_data=home_data,
            )
            LOG.debug("Processed home: %s (%s)", self.homes[home_id].name, home_id)

        LOG.info("Topology updated. Found %d homes.", len(self.homes))

    async def async_get_home_status(
        self,
        home_id: str,
        device_types: list[str] | None = None,
    ) -> dict[str, Any]:
        """Retrieve the current status of modules for a specific home.

        Calls the `/syncapi/v1/homestatus` endpoint. Note that this method
        currently returns the raw API response dictionary. Further processing
        to update `Module` objects might be added in the future.

        Args:
            home_id (str): The ID of the home for which to retrieve the status.
                           Must exist in `self.homes`.
            device_types (Optional[list[str]]): A list of device type strings to
                filter the results. Defaults to `DEFAULT_DEVICE_TYPES`.

        Returns:
            dict[str, Any]: The raw JSON response from the API containing the
                            home status, or an empty dictionary if the home_id
                            is not found (an error is logged).

        Raises:
            AuthError: If obtaining an access token fails.
            ApiError: If the API call fails.

        """
        if home_id not in self.homes:
            LOG.error("Home ID %s not found in known homes.", home_id)
            # Or raise a specific error? For now, return empty dict.
            return {}

        if device_types is None:
            device_types = DEFAULT_DEVICE_TYPES

        payload = {
            "app_type": "app_camera",  # Keep this specific type for pybticino
            "app_version": self._app_version,
            "home_id": home_id,
            "device_types": device_types,
        }

        # Note: Using the specific HOMESTATUS_ENDPOINT from pybticino's const.py
        status_data = await self._async_post_api_request(
            endpoint=HOMESTATUS_ENDPOINT,
            json_data=payload,
        )

        # For now, just return the raw data
        LOG.debug("Raw status data for home %s: %s", home_id, status_data)
        return status_data

    async def async_set_module_state(
        self,
        home_id: str,
        module_id: str,
        state: dict,
        timezone: str | None = None,  # Keep timezone optionality
        bridge_id: str | None = None,  # Keep bridge_id optionality
    ) -> dict[str, Any]:
        """Set the state of a specific module.

        Calls the `/syncapi/v1/setstate` endpoint to change the state of a module
        (e.g., unlock a door lock, turn on a light).

        Args:
            home_id (str): The ID of the home containing the module.
            module_id (str): The ID of the module to control.
            state (dict): A dictionary representing the desired state change.
                          The keys and values depend on the module type
                          (e.g., `{'lock': False}` for unlocking).
            timezone (Optional[str]): The timezone string (e.g., 'Europe/Rome').
                                      Recommended to include based on API logs.
            bridge_id (Optional[str]): The ID of the bridge module, if the target
                                       module is connected via a bridge.

        Returns:
            dict[str, Any]: The raw JSON response from the API, typically
                            indicating the status of the operation.

        Raises:
            ValueError: If the provided `home_id` is not found in `self.homes`.
            AuthError: If obtaining an access token fails.
            ApiError: If the API call fails.

        """
        # Basic validation
        if home_id not in self.homes:
            LOG.error("Home ID %s not found for setting state.", home_id)
            err_msg = f"Home ID {home_id} not found."
            raise ValueError(err_msg)
        # We might not have the module object readily available without processing status first
        # For now, we proceed without checking module existence in self.homes[home_id].modules

        # Construct the module payload part (same logic as original)
        module_payload = {"id": module_id}
        module_payload.update(state)

        if bridge_id and module_id != bridge_id:
            module_payload["bridge"] = bridge_id
            LOG.debug(
                "Adding bridge ID %s to payload for module %s",
                bridge_id,
                module_id,
            )
        elif bridge_id and module_id == bridge_id:
            LOG.debug(
                "Target module %s is the bridge, not adding bridge ID to payload.",
                module_id,
            )
        elif not bridge_id and "-" in module_id and ":" not in module_id:
            LOG.warning(
                "Setting state for potentially bridged module %s without explicit bridge ID. Call might fail.",
                module_id,
            )

        home_payload = {"id": home_id, "modules": [module_payload]}

        if timezone:
            home_payload["timezone"] = timezone
            LOG.debug("Adding timezone %s to home payload", timezone)
        else:
            # Keep the warning from original code
            LOG.warning(
                "Calling setstate without timezone. Call might fail based on logs.",
            )

        payload = {
            "app_type": "app_camera",  # Keep this specific type for pybticino
            "app_version": self._app_version,
            "home": home_payload,
        }

        # Note: Using the specific SETSTATE_ENDPOINT from pybticino's const.py
        result = await self._async_post_api_request(
            endpoint=SETSTATE_ENDPOINT,
            json_data=payload,
        )
        LOG.debug("Set state result: %s", result)
        return result

    async def async_get_events(self, home_id: str, size: int = 30) -> dict[str, Any]:
        """Retrieve the event history for a specific home.

        Calls the `/api/getevents` endpoint to fetch recent events for the home.
        Note that this method currently returns the raw API response dictionary.
        Further processing into `Event` objects might be added later.

        Args:
            home_id (str): The ID of the home for which to retrieve events.
                           Must exist in `self.homes`.
            size (int): The maximum number of events to retrieve. Defaults to 30.

        Returns:
            dict[str, Any]: The raw JSON response from the API containing the
                            events, or an empty dictionary if the home_id is not
                            found (an error is logged).

        Raises:
            AuthError: If obtaining an access token fails.
            ApiError: If the API call fails.

        """
        if home_id not in self.homes:
            LOG.error("Home ID %s not found for getting events.", home_id)
            return {}  # Return empty dict or raise?

        payload = {
            "app_type": "app_camera",  # Keep this specific type for pybticino
            "app_version": self._app_version,
            "home_id": home_id,
            "size": size,
        }

        # Note: Using the specific GETEVENTS_ENDPOINT from pybticino's const.py
        events_data = await self._async_post_api_request(
            endpoint=GETEVENTS_ENDPOINT,
            json_data=payload,
        )

        # For now, just return the raw data
        LOG.debug("Raw events data for home %s: %s", home_id, events_data)
        return events_data

    # --- Add other account-level methods here ---



"""Tests for AsyncAccount API interactions."""

from aioresponses import aioresponses
import pytest

from pybticino import AsyncAccount, AuthHandler
from pybticino.exceptions import ApiError

from .conftest import (
    GETEVENTS_URL,
    HOMESDATA_URL,
    HOMESTATUS_URL,
    MOCK_BRIDGE_ID,
    MOCK_HOME_ID,
    MOCK_HOME_NAME,
    MOCK_LIGHT_ID,
    MOCK_LOCK_ID,
    MOCK_PASSWORD,
    MOCK_USERNAME,
    SETSTATE_URL,
    TOKEN_URL,
    build_events_response,
    build_homesdata_response,
    build_homestatus_response,
    build_token_response,
)


@pytest.fixture
async def authenticated_account():
    """Create an authenticated AsyncAccount with mocked token."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    handler._access_token = "mock-token"
    handler._token_expires_at = 9999999999.0  # Far future
    handler._refresh_token = "mock-refresh"
    account = AsyncAccount(handler)
    yield account
    await handler.close_session()


# --- Instantiation ---


def test_account_requires_auth_handler():
    """Test AsyncAccount rejects non-AuthHandler arguments."""
    with pytest.raises(TypeError, match="must be an instance of AuthHandler"):
        AsyncAccount("not_an_auth_handler")


def test_account_instantiation():
    """Test AsyncAccount can be instantiated with a valid AuthHandler."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    account = AsyncAccount(handler)
    assert account.auth_handler is handler
    assert account.homes == {}
    assert account.user is None


# --- Topology ---


async def test_update_topology(authenticated_account):
    """Test fetching and parsing home topology."""
    account = authenticated_account

    with aioresponses() as m:
        m.post(HOMESDATA_URL, payload=build_homesdata_response())

        await account.async_update_topology()

    assert account.user == MOCK_USERNAME
    assert len(account.homes) == 1
    assert MOCK_HOME_ID in account.homes

    home = account.homes[MOCK_HOME_ID]
    assert home.name == MOCK_HOME_NAME
    assert len(home.modules) == 3

    # Check module types
    module_ids = {m.id for m in home.modules}
    assert MOCK_BRIDGE_ID in module_ids
    assert MOCK_LOCK_ID in module_ids
    assert MOCK_LIGHT_ID in module_ids


async def test_update_topology_empty_homes(authenticated_account):
    """Test topology update when no homes are returned."""
    account = authenticated_account

    with aioresponses() as m:
        m.post(HOMESDATA_URL, payload={"body": {"user": {"email": MOCK_USERNAME}, "homes": []}})

        await account.async_update_topology()

    assert account.user == MOCK_USERNAME
    assert len(account.homes) == 0


async def test_update_topology_api_error(authenticated_account):
    """Test topology update handles API errors."""
    account = authenticated_account

    with aioresponses() as m:
        m.post(HOMESDATA_URL, status=500, body="Internal Server Error")

        with pytest.raises(ApiError):
            await account.async_update_topology()


async def test_update_topology_clears_previous_homes(authenticated_account):
    """Test that topology update replaces previous home data."""
    account = authenticated_account

    with aioresponses() as m:
        # First call
        m.post(HOMESDATA_URL, payload=build_homesdata_response())
        await account.async_update_topology()
        assert len(account.homes) == 1

        # Second call with empty homes
        m.post(HOMESDATA_URL, payload={"body": {"user": {"email": MOCK_USERNAME}, "homes": []}})
        await account.async_update_topology()
        assert len(account.homes) == 0


# --- Home Status ---


async def test_get_home_status(authenticated_account):
    """Test fetching home status."""
    account = authenticated_account

    # First populate homes
    with aioresponses() as m:
        m.post(HOMESDATA_URL, payload=build_homesdata_response())
        await account.async_update_topology()

    with aioresponses() as m:
        m.post(HOMESTATUS_URL, payload=build_homestatus_response())

        status = await account.async_get_home_status(MOCK_HOME_ID)

    assert "body" in status
    modules = status["body"]["home"]["modules"]
    assert len(modules) == 3


async def test_get_home_status_unknown_home(authenticated_account):
    """Test fetching status for unknown home returns empty dict."""
    account = authenticated_account
    result = await account.async_get_home_status("unknown_home_id")
    assert result == {}


async def test_get_home_status_api_error(authenticated_account):
    """Test home status handles API errors."""
    account = authenticated_account

    with aioresponses() as m:
        m.post(HOMESDATA_URL, payload=build_homesdata_response())
        await account.async_update_topology()

    with aioresponses() as m:
        m.post(HOMESTATUS_URL, status=503, body="Service Unavailable")

        with pytest.raises(ApiError):
            await account.async_get_home_status(MOCK_HOME_ID)


# --- Set Module State ---


async def test_set_module_state(authenticated_account):
    """Test setting module state (e.g., unlock door)."""
    account = authenticated_account

    with aioresponses() as m:
        m.post(HOMESDATA_URL, payload=build_homesdata_response())
        await account.async_update_topology()

    with aioresponses() as m:
        m.post(SETSTATE_URL, payload={"status": "ok"})

        result = await account.async_set_module_state(
            home_id=MOCK_HOME_ID,
            module_id=MOCK_LOCK_ID,
            bridge_id=MOCK_BRIDGE_ID,
            state={"lock": False},
            timezone="Europe/Rome",
        )

    assert result["status"] == "ok"


async def test_set_module_state_unknown_home(authenticated_account):
    """Test setting state for unknown home raises ValueError."""
    account = authenticated_account

    with pytest.raises(ValueError, match="not found"):
        await account.async_set_module_state(
            home_id="unknown",
            module_id=MOCK_LOCK_ID,
            state={"lock": False},
        )


async def test_set_module_state_api_error(authenticated_account):
    """Test setting state handles API errors."""
    account = authenticated_account

    with aioresponses() as m:
        m.post(HOMESDATA_URL, payload=build_homesdata_response())
        await account.async_update_topology()

    with aioresponses() as m:
        m.post(SETSTATE_URL, status=500, body="Error")

        with pytest.raises(ApiError):
            await account.async_set_module_state(
                home_id=MOCK_HOME_ID,
                module_id=MOCK_LOCK_ID,
                state={"lock": False},
            )


# --- Get Events ---


async def test_get_events(authenticated_account):
    """Test fetching events for a home."""
    account = authenticated_account

    with aioresponses() as m:
        m.post(HOMESDATA_URL, payload=build_homesdata_response())
        await account.async_update_topology()

    with aioresponses() as m:
        m.post(GETEVENTS_URL, payload=build_events_response())

        events = await account.async_get_events(MOCK_HOME_ID, size=20)

    assert "body" in events
    event_list = events["body"]["home"]["events"]
    assert len(event_list) == 1
    assert event_list[0]["type"] == "call"


async def test_get_events_unknown_home(authenticated_account):
    """Test fetching events for unknown home returns empty dict."""
    account = authenticated_account
    result = await account.async_get_events("unknown_home_id")
    assert result == {}


# --- Integration: Full flow ---


async def test_full_auth_topology_status_flow():
    """Test the full flow: authenticate → topology → status."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        # Auth
        m.post(TOKEN_URL, payload=build_token_response())

        # Topology
        m.post(HOMESDATA_URL, payload=build_homesdata_response())

        # Status
        m.post(HOMESTATUS_URL, payload=build_homestatus_response())

        # Events
        m.post(GETEVENTS_URL, payload=build_events_response())

        # Execute full flow
        account = AsyncAccount(handler)
        await account.async_update_topology()

        assert len(account.homes) == 1
        assert account.user == MOCK_USERNAME

        status = await account.async_get_home_status(MOCK_HOME_ID)
        assert len(status["body"]["home"]["modules"]) == 3

        events = await account.async_get_events(MOCK_HOME_ID)
        assert len(events["body"]["home"]["events"]) == 1

    await handler.close_session()



"""Constants for pybticino."""

import base64
import binascii  # Import for specific exception handling
import logging

_LOGGER = logging.getLogger(__name__)

# Base URL for the Netatmo API
BASE_URL = "https://app.netatmo.net"

# OAuth2 endpoints
TOKEN_ENDPOINT = "/oauth2/token"  # nosec B105

# Client credentials (obfuscated using Base64)
_ENCODED_CLIENT_ID = "bmFfY2xpZW50X2FuZHJvaWRfd2VsY29tZQ=="
_ENCODED_CLIENT_SECRET = "OGFiNTg0ZDYyY2EyYTc3ZTM3Y2NjNmIyYzdlNGYyOWU="  # nosec B105


def get_client_id() -> str:
    """Return the Base64 decoded client ID."""
    try:
        # Ensure correct padding if needed, though standard encoding usually handles it
        return base64.b64decode(_ENCODED_CLIENT_ID).decode("utf-8")
    except binascii.Error, UnicodeDecodeError:
        # Log error or raise a specific exception if decoding fails
        _LOGGER.exception("Error decoding client ID")
        return ""


def get_client_secret() -> str:
    """Return the Base64 decoded client secret."""
    try:
        return base64.b64decode(_ENCODED_CLIENT_SECRET).decode("utf-8")
    except binascii.Error, UnicodeDecodeError:
        _LOGGER.exception("Error decoding client secret")
        return ""


# Default scope (from logs)
DEFAULT_SCOPE = "security_scopes"

# Default App Version (from logs)
DEFAULT_APP_VERSION = "4.1.1.3"
# Default Build Number (from logs User-Agent)
DEFAULT_BUILD_NUMBER = "4401103"
# Default Android Version (from logs User-Agent)
DEFAULT_ANDROID_VERSION = "13"
# Default Device Info (from logs User-Agent)
DEFAULT_DEVICE_INFO = "Google/sdk_gphone64_arm64"
# Default Platform (from logs Subscribe message)
DEFAULT_PLATFORM = "Android"


def build_user_agent(
    app_version: str = DEFAULT_APP_VERSION,
    build_number: str = DEFAULT_BUILD_NUMBER,
    android_version: str = DEFAULT_ANDROID_VERSION,
    device_info: str = DEFAULT_DEVICE_INFO,
) -> str:
    """Build the User-Agent string based on components."""
    # Format based on observed User-Agent:
    # NetatmoApp(Security/v4.1.1.3/4401103) Android(13/Google/sdk_gphone64_arm64)
    return f"NetatmoApp(Security/v{app_version}/{build_number}) Android({android_version}/{device_info})"


# API Endpoints
HOMESDATA_ENDPOINT = "/api/homesdata"
HOMESTATUS_ENDPOINT = "/syncapi/v1/homestatus"
SETSTATE_ENDPOINT = "/syncapi/v1/setstate"
GETEVENTS_ENDPOINT = "/api/getevents"  # Note: Log shows /%2Fapi%2Fgetevents, using decoded path
ADD_PUSH_CONTEXT_ENDPOINT = "/api/addpushcontext"
MODIFY_USER_ENDPOINT = "/api/modifyuser"
GET_HOME_USERS_ENDPOINT = "/api/gethomeusers"
UPDATE_SESSION_ENDPOINT = "/api/updatesession"

# Other hosts / WS URLs
PUSH_WS_URL = "wss://app-ws.netatmo.net/ws/"  # For push notifications (app_camera)
RTC_WS_URL = "wss://app.netatmo.net/appws/"  # For WebRTC signaling (app_security)
NEWS_HOST = "api-news.netatmo.net"



"""Tests for AuthHandler authentication and token management."""

import time

from aioresponses import aioresponses
import pytest

from pybticino import AuthHandler
from pybticino.exceptions import AuthError

from .conftest import (
    MOCK_ACCESS_TOKEN,
    MOCK_PASSWORD,
    MOCK_REFRESH_TOKEN,
    MOCK_USERNAME,
    TOKEN_URL,
    build_token_response,
)

# --- Authentication ---


async def test_authenticate_success():
    """Test successful authentication with username/password."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        m.post(TOKEN_URL, payload=build_token_response())

        token = await handler.get_access_token()

        assert token == MOCK_ACCESS_TOKEN
        assert handler._refresh_token == MOCK_REFRESH_TOKEN
        assert handler._token_expires_at is not None
        assert not handler._is_token_expired()

    await handler.close_session()


async def test_authenticate_invalid_credentials():
    """Test authentication failure with invalid credentials."""
    handler = AuthHandler(MOCK_USERNAME, "wrong_password")

    with aioresponses() as m:
        m.post(
            TOKEN_URL,
            status=400,
            payload={"error": "invalid_grant", "error_description": "Bad credentials"},
        )

        with pytest.raises(AuthError, match="Invalid credentials"):
            await handler.get_access_token()

    await handler.close_session()


async def test_authenticate_server_error():
    """Test authentication failure on server error."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        m.post(TOKEN_URL, status=500, body="Internal Server Error")

        with pytest.raises(AuthError):
            await handler.get_access_token()

    await handler.close_session()


async def test_authenticate_missing_tokens_in_response():
    """Test authentication failure when response lacks tokens."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        m.post(TOKEN_URL, payload={"status": "ok"})

        with pytest.raises(AuthError, match="Missing tokens"):
            await handler.get_access_token()

    await handler.close_session()


async def test_authenticate_network_error():
    """Test authentication failure on network error."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        from aiohttp import ClientConnectionError

        m.post(TOKEN_URL, exception=ClientConnectionError("Connection refused"))

        with pytest.raises(AuthError, match="Request error"):
            await handler.get_access_token()

    await handler.close_session()


# --- Token Refresh ---


async def test_token_refresh_on_expiry():
    """Test that expired token triggers a refresh."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        # Initial auth
        m.post(TOKEN_URL, payload=build_token_response())
        await handler.get_access_token()

        # Simulate token expiry
        handler._token_expires_at = time.time() - 10

        # Refresh
        new_token = "refreshed-token-new"
        m.post(
            TOKEN_URL,
            payload=build_token_response(access_token=new_token),
        )
        token = await handler.get_access_token()

        assert token == new_token

    await handler.close_session()


async def test_token_refresh_failure_falls_back_to_full_auth():
    """Test that refresh failure triggers full re-authentication."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        # Initial auth
        m.post(TOKEN_URL, payload=build_token_response())
        await handler.get_access_token()

        # Simulate token expiry
        handler._token_expires_at = time.time() - 10

        # Refresh fails
        m.post(TOKEN_URL, status=400, payload={"error": "invalid_grant"})

        # Full re-auth succeeds
        reauth_token = "reauth-token"
        m.post(TOKEN_URL, payload=build_token_response(access_token=reauth_token))

        token = await handler.get_access_token()
        assert token == reauth_token

    await handler.close_session()


async def test_token_not_expired_returns_cached():
    """Test that a valid token is returned without re-authenticating."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        m.post(TOKEN_URL, payload=build_token_response())
        token1 = await handler.get_access_token()

        # Second call should NOT make another HTTP request
        token2 = await handler.get_access_token()
        assert token1 == token2

    await handler.close_session()


# --- Token Expiry Check ---


def test_is_token_expired_no_expiry():
    """Test token is considered expired when no expiry is set."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    assert handler._is_token_expired() is True


def test_is_token_expired_within_buffer():
    """Test token is considered expired within 60s buffer."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    handler._token_expires_at = time.time() + 30  # Expires in 30s (within 60s buffer)
    assert handler._is_token_expired() is True


def test_is_token_not_expired():
    """Test token is valid when well within expiry."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    handler._token_expires_at = time.time() + 3600
    assert handler._is_token_expired() is False


# --- Session Management ---


async def test_close_managed_session():
    """Test closing an internally managed session."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    with aioresponses() as m:
        m.post(TOKEN_URL, payload=build_token_response())
        await handler.get_access_token()

    assert handler._managed_session is True
    await handler.close_session()
    assert handler._session is None


async def test_external_session_not_closed():
    """Test that externally provided session is not closed by handler."""
    import aiohttp

    external_session = aiohttp.ClientSession()
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD, session=external_session)

    assert handler._managed_session is False
    await handler.close_session()
    # Session should still be open
    assert not external_session.closed
    await external_session.close()



"""Data models for pybticino."""

from dataclasses import dataclass, field
from typing import Any

# Placeholder models - To be refined based on API responses


@dataclass
class Module:
    """Represents a single BTicino/Netatmo device or module.

    Attributes:
        id (str): The unique identifier of the module (MAC address or similar).
        name (str): The user-defined name of the module.
        type (str): The type identifier of the module (e.g., 'BNDL' for door lock).
        bridge (Optional[str]): The ID of the bridge module this module is connected to,
                                 if applicable. None for main/bridge modules.
        raw_data (dict[str, Any]): The raw dictionary representation of the module
                                   as received from the API. Useful for accessing
                                   less common or undocumented attributes.

    """

    id: str
    name: str
    type: str
    bridge: str | None = None
    # Add other common attributes observed in homesdata/homestatus
    raw_data: dict[str, Any] = field(default_factory=dict)  # Store the raw dictionary


@dataclass
class Home:
    """Represents a BTicino/Netatmo home installation.

    Attributes:
        id (str): The unique identifier of the home.
        name (str): The user-defined name of the home.
        modules (list[Module]): A list of Module objects belonging to this home.
        raw_data (dict[str, Any]): The raw dictionary representation of the home
                                   as received from the API.

    """

    id: str
    name: str
    modules: list[Module]
    # Add other attributes from homesdata/homestatus if needed
    raw_data: dict[str, Any] = field(default_factory=dict)  # Store the raw dictionary


@dataclass
class Event:
    """Represents an event recorded by the BTicino/Netatmo system.

    Attributes:
        id (str): The unique identifier of the event.
        type (str): The type of the event (e.g., 'doorbell_ring', 'person_seen').
        time (int): The timestamp (Unix epoch) when the event occurred.
        raw_data (dict[str, Any]): The raw dictionary representation of the event
                                   as received from the API. Contains additional
                                   details specific to the event type.

    """

    id: str
    type: str
    time: int
    # Add other event attributes
    raw_data: dict[str, Any] = field(default_factory=dict)  # Store the raw dictionary



"""Tests for WebsocketClient."""

import asyncio
import json
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from pybticino import AuthHandler, WebsocketClient
from pybticino.exceptions import PyBticinoException

from .conftest import MOCK_ACCESS_TOKEN, MOCK_PASSWORD, MOCK_USERNAME


class AsyncIterFromList:
    """Async iterator wrapper for a list of items."""

    def __init__(self, items):
        self._items = iter(items)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self._items)
        except StopIteration:
            raise StopAsyncIteration from None


# --- Instantiation ---


def test_websocket_requires_auth_handler():
    """Test WebsocketClient rejects non-AuthHandler."""

    async def cb(msg):
        pass

    with pytest.raises(TypeError, match="must be an instance of AuthHandler"):
        WebsocketClient("not_handler", cb)


def test_websocket_requires_async_callback():
    """Test WebsocketClient rejects non-async callback."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    def sync_cb(msg):
        pass

    with pytest.raises(TypeError, match="must be an async function"):
        WebsocketClient(handler, sync_cb)


def test_websocket_instantiation():
    """Test WebsocketClient can be instantiated."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    async def cb(msg):
        pass

    client = WebsocketClient(handler, cb)
    assert client._auth_handler is handler
    assert client._is_running is False
    assert client.get_listener_task() is None


# --- Connect / Disconnect ---


async def test_connect_and_subscribe():
    """Test connect establishes connection and subscribes."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    handler._access_token = MOCK_ACCESS_TOKEN
    handler._token_expires_at = 9999999999.0

    received_messages = []

    async def cb(msg):
        received_messages.append(msg)

    client = WebsocketClient(handler, cb)

    mock_ws = AsyncMock()
    mock_ws.state = MagicMock()
    mock_ws.send = AsyncMock()
    mock_ws.recv = AsyncMock(return_value=json.dumps({"status": "ok"}))

    with patch("pybticino.websocket.websockets.connect", new_callable=AsyncMock, return_value=mock_ws):
        await client.connect()

    assert client._is_running is True
    assert client._websocket is mock_ws
    mock_ws.send.assert_called_once()

    # Check subscription payload
    sent_payload = json.loads(mock_ws.send.call_args[0][0])
    assert sent_payload["access_token"] == MOCK_ACCESS_TOKEN
    assert sent_payload["action"] == "Subscribe"

    # Cleanup
    await client.disconnect()


async def test_connect_failure_resets_state():
    """Test that failed connection resets running state."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    handler._access_token = MOCK_ACCESS_TOKEN
    handler._token_expires_at = 9999999999.0

    async def cb(msg):
        pass

    client = WebsocketClient(handler, cb)

    with (
        patch(
            "pybticino.websocket.websockets.connect",
            new_callable=AsyncMock,
            side_effect=OSError("Connection refused"),
        ),
        pytest.raises(PyBticinoException, match="connection/subscription failed"),
    ):
        await client.connect()

    assert client._is_running is False
    assert client._websocket is None


async def test_disconnect_when_not_connected():
    """Test disconnect is safe when not connected."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)

    async def cb(msg):
        pass

    client = WebsocketClient(handler, cb)
    await client.disconnect()  # Should not raise


async def test_subscribe_failure_raises():
    """Test that subscription failure raises PyBticinoException."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    handler._access_token = MOCK_ACCESS_TOKEN
    handler._token_expires_at = 9999999999.0

    async def cb(msg):
        pass

    client = WebsocketClient(handler, cb)

    mock_ws = AsyncMock()
    mock_ws.state = MagicMock()
    mock_ws.send = AsyncMock()
    mock_ws.recv = AsyncMock(return_value=json.dumps({"status": "error", "msg": "bad token"}))
    mock_ws.close = AsyncMock()

    with (
        patch("pybticino.websocket.websockets.connect", new_callable=AsyncMock, return_value=mock_ws),
        pytest.raises(PyBticinoException, match="subscription failed"),
    ):
        await client.connect()

    assert client._is_running is False


async def test_listener_receives_messages():
    """Test that the listener invokes the callback for received messages."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    handler._access_token = MOCK_ACCESS_TOKEN
    handler._token_expires_at = 9999999999.0

    received = []

    async def cb(msg):
        received.append(msg)

    client = WebsocketClient(handler, cb)

    test_message = {"type": "push", "data": {"event": "call"}}

    mock_ws = AsyncMock()
    mock_ws.state = MagicMock()
    mock_ws.send = AsyncMock()
    mock_ws.recv = AsyncMock(return_value=json.dumps({"status": "ok"}))
    mock_ws.close = AsyncMock()

    # Make the websocket iterable: yield one message then stop
    mock_ws.__aiter__ = MagicMock(return_value=AsyncIterFromList([json.dumps(test_message)]))

    with patch("pybticino.websocket.websockets.connect", new_callable=AsyncMock, return_value=mock_ws):
        await client.connect()

        # Wait for listener to process
        if client._listener_task:
            await asyncio.wait_for(client._listener_task, timeout=2.0)

    assert len(received) == 1
    assert received[0] == test_message

    await client.disconnect()


async def test_listener_handles_invalid_json():
    """Test that the listener handles non-JSON messages gracefully."""
    handler = AuthHandler(MOCK_USERNAME, MOCK_PASSWORD)
    handler._access_token = MOCK_ACCESS_TOKEN
    handler._token_expires_at = 9999999999.0

    received = []

    async def cb(msg):
        received.append(msg)

    client = WebsocketClient(handler, cb)

    mock_ws = AsyncMock()
    mock_ws.state = MagicMock()
    mock_ws.send = AsyncMock()
    mock_ws.recv = AsyncMock(return_value=json.dumps({"status": "ok"}))
    mock_ws.close = AsyncMock()

    # Yield invalid JSON then valid message
    mock_ws.__aiter__ = MagicMock(return_value=AsyncIterFromList(["not json at all", json.dumps({"valid": True})]))

    with patch("pybticino.websocket.websockets.connect", new_callable=AsyncMock, return_value=mock_ws):
        await client.connect()
        if client._listener_task:
            await asyncio.wait_for(client._listener_task, timeout=2.0)

    # Only the valid message should have been received
    assert len(received) == 1
    assert received[0] == {"valid": True}

    await client.disconnect()



"""Async Authentication handler for pybticino."""

import logging
import time

import aiohttp

from .const import (
    BASE_URL,
    DEFAULT_APP_VERSION,
    DEFAULT_SCOPE,
    TOKEN_ENDPOINT,
    get_client_id,
    get_client_secret,
)
from .exceptions import ApiError, AuthError

_LOGGER = logging.getLogger(__name__)


class AuthHandler:
    """Handles asynchronous authentication and token management for the BTicino API.

    This class manages the OAuth2 password grant flow to obtain access and
    refresh tokens. It automatically handles token expiration and refreshing.
    It uses an `aiohttp.ClientSession` for making HTTP requests.

    Attributes:
        _username (str): The user's email address for authentication.
        _password (str): The user's password for authentication.
        _client_id (str): The client ID for the API application.
        _client_secret (str): The client secret for the API application.
        _scope (str): The requested scope for the access token.
        _app_version (str): The application version string sent during auth.
        _access_token (Optional[str]): The current access token.
        _refresh_token (Optional[str]): The current refresh token.
        _token_expires_at (Optional[float]): The timestamp when the access token expires.
        _session (Optional[aiohttp.ClientSession]): The HTTP client session.
        _managed_session (bool): True if the session was created internally and should
                                 be closed by this handler, False otherwise.

    """

    def __init__(
        self,
        username: str,
        password: str,
        scope: str = DEFAULT_SCOPE,
        app_version: str = DEFAULT_APP_VERSION,
        session: aiohttp.ClientSession | None = None,
    ) -> None:
        """Initialize the asynchronous authentication handler.

        Args:
            username (str): The user's email address (BTicino/Netatmo account).
            password (str): The user's password.
            scope (str): The requested OAuth scope. Defaults to `DEFAULT_SCOPE`.
            app_version (str): The application version string. Defaults to
                               `DEFAULT_APP_VERSION`.
            session (Optional[aiohttp.ClientSession]): An optional existing
                `aiohttp.ClientSession` to use for requests. If None, a new
                session will be created and managed internally.

        """
        self._username = username
        self._password = password
        self._client_id = get_client_id()
        self._client_secret = get_client_secret()
        self._scope = scope
        self._app_version = app_version
        self._access_token: str | None = None
        self._refresh_token: str | None = None
        self._token_expires_at: float | None = None
        # Use provided session or create a new one
        self._session = session
        self._managed_session = session is None  # Flag to know if we should close the session

    async def _get_session(self) -> aiohttp.ClientSession:
        """Get the current `aiohttp.ClientSession` or create one if needed.

        If a session was provided during initialization, it will be returned.
        Otherwise, if no session exists or the existing one is closed, a new
        session is created and marked as managed by this handler.

        Returns:
            aiohttp.ClientSession: The active client session.

        """
        if self._session is None or self._session.closed:
            _LOGGER.debug("Creating new aiohttp ClientSession for AuthHandler.")
            self._session = aiohttp.ClientSession()
            self._managed_session = True  # We created it, so we manage it
        return self._session

    async def close_session(self) -> None:
        """Close the `aiohttp.ClientSession` if it was created internally.

        If the session was provided externally during initialization, this method
        does nothing.
        """
        if self._session and not self._session.closed and self._managed_session:
            await self._session.close()
            self._session = None
            _LOGGER.debug("Managed aiohttp session closed by AuthHandler.")
        elif self._session and not self._managed_session:
            _LOGGER.debug("Session provided externally, not closing.")

    def _is_token_expired(self) -> bool:
        """Check if the current access token is expired or nearing expiration.

        Considers the token expired if the expiration timestamp is not set or
        if the current time is within 60 seconds of the expiration time.

        Returns:
            bool: True if the token is expired or nearing expiration, False otherwise.

        """
        if not self._token_expires_at:
            return True
        return time.time() >= (self._token_expires_at - 60)

    async def get_access_token(self) -> str:
        """Return a valid access token, automatically authenticating or refreshing.

        This method checks if the current token is valid. If it's expired or
        nearing expiration, it attempts to refresh it using the refresh token.
        If refreshing fails or no refresh token is available, it performs a full
        authentication using the username and password.

        Returns:
            str: A valid access token.

        Raises:
            AuthError: If authentication or token refresh fails persistently.

        """
        if self._is_token_expired():
            _LOGGER.debug("Token is expired or nearing expiration.")
            if self._refresh_token:
                try:
                    await self._refresh_access_token()
                except (AuthError, ApiError) as e:
                    _LOGGER.warning(
                        "Token refresh failed (%s), attempting full authentication.",
                        type(e).__name__,
                    )
                    await self.authenticate()  # Fallback to full auth
            else:
                _LOGGER.debug(
                    "No refresh token available, performing full authentication.",
                )
                await self.authenticate()
        elif not self._access_token:
            _LOGGER.debug("No access token found, performing full authentication.")
            await self.authenticate()

        if not self._access_token:
            # Should not happen if authenticate/refresh worked, but safety check
            err_msg = "Failed to obtain a valid access token."
            raise AuthError(err_msg)

        return self._access_token

    async def authenticate(self) -> None:
        """Perform the initial authentication using username and password.

        This method makes a POST request to the token endpoint with the
        'password' grant type to obtain the initial access and refresh tokens.
        It stores the tokens and their expiration time internally.

        Raises:
            AuthError: If authentication fails due to invalid credentials,
                       network errors, timeouts, or unexpected issues.
            ApiError: If the API returns a non-success status code other than
                      common authentication failures handled by AuthError.

        """
        url = BASE_URL + TOKEN_ENDPOINT
        payload = {
            "grant_type": "password",
            "username": self._username,
            "password": self._password,
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "scope": self._scope,
            "app_version": self._app_version,
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        session = await self._get_session()

        try:
            _LOGGER.debug("Requesting token from %s", url)
            async with session.post(
                url,
                data=payload,
                headers=headers,
                timeout=10,
            ) as response:
                if response.status >= 400:
                    error_text = await response.text()
                    _LOGGER.error(
                        "HTTP error %s during authentication: %s",
                        response.status,
                        error_text,
                    )
                    try:
                        error_data = await response.json()
                        if response.status == 400 and error_data.get("error") == "invalid_grant":
                            err_msg = "Authentication failed: Invalid credentials or grant"
                            raise AuthError(err_msg)
                        # Raise generic ApiError for other 4xx/5xx based on JSON if possible
                        raise ApiError(
                            response.status,
                            error_data.get("error", error_text),
                        )
                    except (
                        aiohttp.ContentTypeError,
                        ValueError,
                    ) as parse_err:  # Handle non-JSON errors
                        raise ApiError(response.status, error_text) from parse_err

                token_data = await response.json()
                _LOGGER.debug("Token response received: %s", token_data)

                if "access_token" not in token_data or "refresh_token" not in token_data:
                    err_msg = "Authentication failed: Missing tokens in response"
                    raise AuthError(err_msg)

                self._access_token = token_data["access_token"]
                self._refresh_token = token_data["refresh_token"]
                expires_in = token_data.get("expires_in")
                if expires_in:
                    self._token_expires_at = time.time() + int(expires_in)
                    _LOGGER.debug("Token expires at: %s", self._token_expires_at)
                else:
                    self._token_expires_at = None
                    _LOGGER.warning("No 'expires_in' found in token response.")
                _LOGGER.info("Authentication successful. Access token obtained.")

        except aiohttp.ClientError as req_err:
            _LOGGER.exception("Request error during authentication")
            err_msg = f"Authentication failed: Request error - {req_err}"
            raise AuthError(err_msg) from req_err
        except TimeoutError as timeout_err:
            _LOGGER.exception("Timeout during authentication request")
            err_msg = "Authentication failed: Request timed out"
            raise AuthError(err_msg) from timeout_err
        except Exception as e:
            _LOGGER.exception("Unexpected error during authentication")
            err_msg = f"Authentication failed: Unexpected error - {e}"
            raise AuthError(err_msg) from e

    async def _refresh_access_token(self) -> None:
        """Refresh the access token using the stored refresh token.

        This method makes a POST request to the token endpoint with the
        'refresh_token' grant type. It updates the internal access token,
        refresh token (if provided in the response), and expiration time.

        Raises:
            AuthError: If no refresh token is available, if the refresh attempt
                       fails due to an invalid grant or request, network errors,
                       timeouts, or unexpected issues.
            ApiError: If the API returns a non-success status code other than
                      common refresh failures handled by AuthError.

        """
        if not self._refresh_token:
            err_msg = "Cannot refresh token: No refresh token available."
            raise AuthError(err_msg)

        url = BASE_URL + TOKEN_ENDPOINT
        payload = {
            "grant_type": "refresh_token",
            "refresh_token": self._refresh_token,
            "client_id": self._client_id,
            "client_secret": self._client_secret,
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        session = await self._get_session()

        try:
            _LOGGER.info("Refreshing access token...")
            async with session.post(
                url,
                data=payload,
                headers=headers,
                timeout=10,
            ) as response:
                if response.status >= 400:
                    error_text = await response.text()
                    _LOGGER.error(
                        "HTTP error %s during token refresh: %s",
                        response.status,
                        error_text,
                    )
                    try:
                        error_data = await response.json()

                        # Only clear tokens for actual auth-related failures
                        if response.status == 400 and error_data.get("error") in [
                            "invalid_grant",
                            "invalid_request",
                        ]:
                            # Clear tokens on persistent refresh failure (like invalid grant)
                            self._access_token = None
                            self._refresh_token = None
                            self._token_expires_at = None
                            err_msg = f"Token refresh failed: {error_data.get('error')}"
                            raise AuthError(err_msg)

                        # For server errors (500) or other errors, keep refresh token
                        if response.status >= 500:
                            # Server error - don't clear refresh token, raise AuthError to trigger fallback
                            _LOGGER.warning(
                                "Server error during token refresh (code %s), keeping refresh token for retry",
                                error_data.get("error", {}).get("code", "unknown"),
                            )
                            err_msg = f"Server error during token refresh: {error_data.get('error', error_text)}"
                            raise AuthError(err_msg)

                        # Other 4xx client errors - keep refresh token but raise ApiError
                        _LOGGER.warning(
                            "Client error during token refresh (%s), keeping refresh token", response.status
                        )
                        raise ApiError(
                            response.status,
                            error_data.get("error", error_text),
                        )
                    except (aiohttp.ContentTypeError, ValueError) as parse_err:
                        # Can't parse JSON response - treat as server error if 5xx, auth error if 4xx
                        if response.status >= 500:
                            err_msg = f"Server error during token refresh: HTTP {response.status} - {error_text}"
                            raise AuthError(err_msg) from parse_err
                        elif response.status == 400 or response.status == 401:
                            # Likely auth failure if we can't parse 400/401 response
                            self._access_token = None
                            self._refresh_token = None
                            self._token_expires_at = None
                            err_msg = f"Token refresh failed: HTTP {response.status} - {error_text}"
                            raise AuthError(err_msg) from parse_err
                        else:
                            # Other client errors - keep refresh token
                            err_msg = f"Token refresh failed: HTTP {response.status} - {error_text}"
                            raise AuthError(err_msg) from parse_err

                token_data = await response.json()
                _LOGGER.debug("Token refresh response received: %s", token_data)

                if "access_token" not in token_data:
                    self._access_token = None
                    self._refresh_token = None
                    self._token_expires_at = None
                    err_msg = "Token refresh failed: Missing access token in response"
                    raise AuthError(err_msg)

                self._access_token = token_data["access_token"]
                self._refresh_token = token_data.get(
                    "refresh_token",
                    self._refresh_token,
                )
                expires_in = token_data.get("expires_in")
                if expires_in:
                    self._token_expires_at = time.time() + int(expires_in)
                    _LOGGER.debug(
                        "Refreshed token expires at: %s",
                        self._token_expires_at,
                    )
                else:
                    self._token_expires_at = None
                    _LOGGER.warning("No 'expires_in' found in token refresh response.")
                _LOGGER.info("Access token refreshed successfully.")

        except aiohttp.ClientError as req_err:
            _LOGGER.exception("Request error during token refresh")
            # Don't clear tokens on temporary network errors
            err_msg = f"Token refresh failed: Request error - {req_err}"
            raise AuthError(err_msg) from req_err
        except TimeoutError as timeout_err:
            _LOGGER.exception("Timeout during token refresh request")
            err_msg = "Token refresh failed: Request timed out"
            raise AuthError(err_msg) from timeout_err
        except Exception as e:
            _LOGGER.exception("Unexpected error during token refresh")
            # Don't clear tokens on unknown errors
            err_msg = f"Token refresh failed: Unexpected error - {e}"
            raise AuthError(err_msg) from e



#!/usr/bin/env python3

"""Example script to unlock a specific door lock (BNDL module) using the async pybticino library.

Reads credentials (BTICINOUSER, BTICINOPASSWORD) and the target home ID
(BTICINOHOMEID) from environment variables.

Requires the target door lock module ID as a command-line argument.

Usage:
  export BTICINOUSER="your_email"
  export BTICINOPASSWORD="your_password"
  export BTICINOHOMEID="your_home_id"
  python3 unlock_door.py <door_lock_module_id>
"""

import argparse
import asyncio
import logging
import os
import sys
from typing import Optional  # Import Optional

# Import new async classes
from pybticino import ApiError, AsyncAccount, AuthError, AuthHandler, Module

# --- Configuration ---
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
)

# Read credentials and home ID from environment variables
USERNAME = os.getenv("BTICINOUSER")
PASSWORD = os.getenv("BTICINOPASSWORD")
HOME_ID = os.getenv("BTICINOHOMEID")

if not all([USERNAME, PASSWORD, HOME_ID]):
    logging.error(
        "Please set BTICINOUSER, BTICINOPASSWORD, and BTICINOHOMEID environment variables.",
    )
    sys.exit(1)

# --- Argument Parsing ---
parser = argparse.ArgumentParser(description="Unlock a BTicino/Netatmo door lock.")
parser.add_argument(
    "module_id",
    help="The ID of the BNDL module (door lock) to unlock.",
)
args = parser.parse_args()
MODULE_ID_TO_UNLOCK = args.module_id


# --- Main Async Function ---
async def main() -> None:
    """Run the async unlock script."""
    logging.info(
        "Attempting to unlock module ID: %s in home ID: %s",
        MODULE_ID_TO_UNLOCK,
        HOME_ID,
    )
    auth = None  # Initialize for finally block
    # Ensure credentials and HOME_ID are not None before proceeding
    if not USERNAME or not PASSWORD or not HOME_ID:
        logging.error(
            "USERNAME, PASSWORD, or HOME_ID is None. Check environment variables.",
        )
        sys.exit(1)
    try:
        # 1. Create AuthHandler and AsyncAccount
        logging.info("Creating AuthHandler and AsyncAccount...")
        auth = AuthHandler(USERNAME, PASSWORD)
        account = AsyncAccount(auth)
        logging.info("Authentication will occur on first API call.")

        # 2. Get Home Data to find timezone and bridge ID
        logging.info(
            "Fetching topology for home %s to get timezone and bridge ID...",
            HOME_ID,
        )
        # This call implicitly authenticates if needed
        await account.async_update_topology()

        home_obj = account.homes.get(HOME_ID)

        if not home_obj:
            logging.error(
                "Could not find home with ID %s after topology update.",
                HOME_ID,
            )
            sys.exit(1)

        home_timezone = home_obj.raw_data.get("timezone")  # Access raw_data for now
        bridge_id = None
        module_details: Optional[Module] = None  # Use Module model

        for module in home_obj.modules:
            if module.type == "BNC1":  # Assuming BNC1 is the bridge
                bridge_id = module.id
            if module.id == MODULE_ID_TO_UNLOCK:
                module_details = module

        if not module_details:
            logging.error(
                "Module %s not found in home %s.",
                MODULE_ID_TO_UNLOCK,
                HOME_ID,
            )
            sys.exit(1)

        if module_details.type != "BNDL":
            logging.error(
                "Module %s is not a Door Lock (BNDL), it is type '%s'. Aborting.",
                MODULE_ID_TO_UNLOCK,
                module_details.type,
            )
            sys.exit(1)

        if not home_timezone:
            logging.warning(
                "Could not determine timezone for the home. Proceeding without it, but this might fail.",
            )
        if not bridge_id:
            logging.error(
                "Could not determine bridge ID (BNC1) for the home. Cannot send command to bridged module.",
            )
            sys.exit(1)

        logging.info("Found target module: Name='%s', Type='BNDL'", module_details.name)
        logging.info("Using Bridge ID: %s", bridge_id)
        logging.info("Using Timezone: %s", home_timezone)

        # 3. Call async_set_module_state to unlock
        logging.info(
            "Sending unlock command (lock=False) to module %s...",
            MODULE_ID_TO_UNLOCK,
        )
        result = await account.async_set_module_state(
            home_id=HOME_ID,
            module_id=MODULE_ID_TO_UNLOCK,
            state={"lock": False},  # The command to unlock
            timezone=home_timezone,
            bridge_id=bridge_id,
        )
        logging.info("Set state command sent. Result: %s", result)
        logging.info(
            "Door lock %s should be unlocked (it might re-lock automatically).",
            MODULE_ID_TO_UNLOCK,
        )

    except AuthError:
        logging.exception("Authentication Error")
    except ApiError as e:
        logging.exception(
            "API Error: Status=%s, Message=%s",
            e.status_code,
            e.error_message,
        )
    except Exception:
        logging.exception("An unexpected error occurred")
    finally:
        # Ensure the session is closed
        if auth:
            await auth.close_session()
            logging.info("Auth session closed.")


if __name__ == "__main__":
    asyncio.run(main())



"""Python library for interacting with the BTicino/Netatmo API."""

# Import main classes for easier access
from .account import AsyncAccount
from .auth import AuthHandler

# Import exceptions for easier handling
from .exceptions import ApiError, AuthError, PyBticinoException
from .models import Event, Home, Module  # Expose models
from .websocket import WebsocketClient

# Define package version (consider using importlib.metadata in the future)
__version__ = "0.1.0"  # Update version to reflect refactoring

# Define what gets imported with 'from pybticino import *'
__all__ = [
    "ApiError",
    "AsyncAccount",
    "AuthError",
    "AuthHandler",
    "Event",
    "Home",
    "Module",
    "PyBticinoException",
    "WebsocketClient",
    "__version__",
]



"""Tests for pybticino data models."""

from pybticino.models import Event, Home, Module


def test_module_defaults():
    """Test Module dataclass defaults."""
    mod = Module(id="m1", name="Test", type="BNDL")
    assert mod.id == "m1"
    assert mod.bridge is None
    assert mod.raw_data == {}


def test_module_with_raw_data():
    """Test Module with raw_data preserves all fields."""
    raw = {"id": "m1", "name": "Test", "type": "BNDL", "firmware": "1.0", "custom_field": 42}
    mod = Module(id="m1", name="Test", type="BNDL", raw_data=raw)
    assert mod.raw_data["firmware"] == "1.0"
    assert mod.raw_data["custom_field"] == 42


def test_home_with_modules():
    """Test Home with a list of modules."""
    m1 = Module(id="m1", name="Lock", type="BNDL")
    m2 = Module(id="m2", name="Light", type="BNSL")
    home = Home(id="h1", name="Casa", modules=[m1, m2])
    assert len(home.modules) == 2
    assert home.modules[0].id == "m1"


def test_home_empty_modules():
    """Test Home with no modules."""
    home = Home(id="h1", name="Empty Home", modules=[])
    assert home.modules == []
    assert home.raw_data == {}


def test_event_dataclass():
    """Test Event dataclass."""
    raw = {"id": "e1", "type": "call", "time": 1700000000, "module_id": "ext1"}
    evt = Event(id="e1", type="call", time=1700000000, raw_data=raw)
    assert evt.time == 1700000000
    assert evt.raw_data["module_id"] == "ext1"


def test_module_equality():
    """Test that two modules with same data are equal."""
    m1 = Module(id="m1", name="Lock", type="BNDL")
    m2 = Module(id="m1", name="Lock", type="BNDL")
    assert m1 == m2


def test_module_inequality():
    """Test that modules with different IDs are not equal."""
    m1 = Module(id="m1", name="Lock", type="BNDL")
    m2 = Module(id="m2", name="Lock", type="BNDL")
    assert m1 != m2
